/* The QDISK program makes IDEDOS partitions on harddisks and .HDF files.*/

//040430 added 32MB CPM22 partition option

#include <stdio.h>
#include <sys/types.h>
#include "idedos.h"

#ifdef ZXVGS
#include <zxvgs.h>
#endif

#ifdef SPECTRUM
// in ZX Spectrum the code must be behind DSKHNDLRs
#pragma output myzorg=40000
#define TAPE_TRIES 10
#endif

#ifdef CPM
#include <fcntl.h>
#endif

#ifdef SMALL_C
#include "qqz80.c"
#endif


#define MAXPART 504
#define MAXFREE 32

static u8_t sector[512]; //sector buffer
static int sector_c=-1,sector_h=-1,sector_s=-1; //the sector in buffer
static part_t partdata; //partition data

static int cylinders=1,heads=2,sectors=1; //disk parameters
static int records; //number of RECORDs in B-DOS
static int idehead=1; //head where IDEDOS starts
static int partmax=1; //last partition number
static u8_t partition[MAXPART]; //partition types

//set of free areas
static struct free_t
 {//free area description
  int	partno;	//partition number
  long	from;	//start absolute head
  long	size;	//number of heads
 } free[MAXFREE]; //32 should be enough
int freecnt=0; //number of found free areas
static long freemax;	//the biggest free place
static long freetotal;	//the total free space
  

//declaration IDEDOS system partition
static part_t idedos; 
static char *plusidedos="PLUSIDEDOS      ";
 
//flags of found structures
static int is_mbr=0;	//MBR of MS-DOS, etc.
static int is_record=0;	//structures of B-DOS for Sam
static int is_idedos=0;	//IDEDOS
static int is_vdt=0;	//VDT of MB-02


#ifndef ZXVGS
#ifdef SPECTRUM
//this one should go into spectrum.lib !!!
int tape_load(char *name, struct zxtapehdr header, void *start, size_t len)
{//loads file from tape into memory
 int tries=TAPE_TRIES; //fail after these
 int mystart=0;
 char buf[10];
 int i;
 for(i=9;i>=0;i--) buf[i]=' ';
 i=0;
 while ((i<10)&&(name[i]!='\0')) //fill buffer with name
  {buf[i]=name[i]; i++;}
 do
  {//load header first
   i=tape_load_block(header,17,0);
   if ((i==0)&&(header->type==3))
    {//then compare name and load rest
     if (name!="")
      for(i=0;i<10;i++) //use strncmp(header->name,buf,10); ???
       if (header->name[i]!=buf[i]) {i=-1; break;}
     if (i>=0)
      i=tape_load_block((start==NULL)?header->address:start,header->length,255);
    }
  }
 while ((i!=0)&&(--tries>0));
 return (i!=0);
}
int loadany(char *b,int adr,int len)
{//loads file from tape into memory
 struct zxtapehdr hdr;
 tape_load(b,hdr,(void*)adr,len);
 return hdr.length; //should return number of loaded bytes...
}
#else
//very slow in CP/M (z88dk)!!!
int loadany(char *b,int adr,int len)
{//loads file into memory; is definded in ZXVGS, must be for others
 FILE *f;
 int i;
 f=fopen(b,"rb");
 if (f==NULL) return 0;
 i=fread((void*)adr,1,len,f);
 fclose(f);
 return i;
}
#endif
#endif

int compare(unsigned char *a,unsigned char *b,int n)
{//compares n bytes of *a and *b strings
 while (*a++==*b++) if (n--==0) break;
 return n==0;
}

void cls()
{
#ifdef SPECTRUM
 putchar(12); //clear screen
#endif
}

void sector_read(int cy,int he,int se)
{//reads sector into buffer
 //printf("Read sector %d/%d/%d to %d.\n",cy,he,se,(int)sector); //for tests
 if ((cy==sector_c)&&(he==sector_h)&&(se==sector_s)) return;
 readsector(0xA0,cy,he,se,sector);
 sector_c=cy;
 sector_h=he;
 sector_s=se;
}

void ide_partition_get(part_t part,int nr)
{//loads partition data from disk into part*
 int s=idehead+nr/8/sectors; //track (temporary)
 int h=s%heads; //heads=2 at least at start
 int c=s/heads;
 s=(nr/8)%sectors+1;
 sector_read(c,h,s);
 memcpy(part,sector+64*(nr%8),sizeof(part_t));
}

void ide_partition_put(part_t part, int nr)
{//saves partition data to disk from part*
 int s=idehead+nr/8/sectors; //track (temporary)
 int h=s%heads;
 int c=s/heads;
 s=(nr/8)%sectors+1;
 sector_read(c,h,s);
 memcpy(sector+64*(nr%8),part,sizeof(part_t));
 writesector(0xA0,c,h,s,sector);
 partition[nr]=part->part_sysinfo.ph_type;
}

void ide_partition_new(part_t p,int type,unsigned char *name,long tracks,long from)
{//fills structure; partition from side "from" and of "tracks" size
 //with tracks==0 the name and type can be changed
 long last;
 int i,j;
 //clear only when dimensions set or unused entry
 if ((tracks>0L)||(type==ptype_unused)) 
  for (i=0;i<sizeof(part_t);i++) p->part_data[i]=0;
 j=0;
 for (i=0;i<16;i++) //pad name with spaces
  if (name[j]>=' ')
   p->part_name[i]=name[j++];
  else
   p->part_name[i]=' ';
 p->part_sysinfo.ph_type=type;
 if (tracks>0L)
  {//set partition dimensions
   last=from+tracks-1;
   p->part_sysinfo.ph_scyl=from/heads;
   p->part_sysinfo.ph_shd=from%heads;
   p->part_sysinfo.ph_ecyl=last/heads;
   p->part_sysinfo.ph_ehd=last%heads;
   p->part_sysinfo.ph_secs=(long)tracks*sectors-1;
  }
}

void free_count(int verbose)
{//calculates free area
 int i;
 freemax=0;
 freetotal=0;
 for (i=0;i<freecnt;i++)
  {
   freetotal+=free[i].size;
   if (free[i].size>freemax) freemax=free[i].size;
  }
 if (verbose) 
  printf(" Largest free area: %ld; total: %ld track(s).\n",freemax,freetotal);
}

int ide_partition_load()
{//loads partition types from disk, collect free space information
 int i; //counter of for
 cylinders=idedos.part_info.sinf.sinf_unit.un_cyls;
 heads    =idedos.part_info.sinf.sinf_unit.un_hds;
 sectors  =idedos.part_info.sinf.sinf_unit.un_spt;
 partmax  =idedos.part_info.sinf.sinf_unit.un_part;
 if ((partmax>MAXPART)||(partmax>8*sectors)) //only 1 head
  {printf("\n\nThere are too many partitions (%d).\n\n",partmax);
   return 0;
  }
 freecnt=0; //free area counter
 if (partmax>1)
  for (i=0;i<=partmax;i++)
   {ide_partition_get(&partdata,i);
    partition[i]=partdata.part_sysinfo.ph_type; //partition type
    if (partition[i]==ptype_free)
     {free[freecnt].partno=i;
      free[freecnt].from=
       (long)partdata.part_sysinfo.ph_scyl*heads+partdata.part_sysinfo.ph_shd;
      free[freecnt].size=
       (long)(partdata.part_sysinfo.ph_ecyl-partdata.part_sysinfo.ph_scyl)*heads+
              partdata.part_sysinfo.ph_ehd -partdata.part_sysinfo.ph_shd+1;
      printf(" Free area #%d: from track %ld, %ld kB.\n",
       i,free[freecnt].from,free[freecnt].size*sectors/2);
      freecnt++;
     }
   }
 free_count(0);  
 return 1;
}

void putfrees()
{//save free area info to disk and clear unused entries
 int i;
 for (i=0;i<freecnt;i++)
  if (free[i].size>0L)
   {ide_partition_new(partdata,ptype_free,"FREE",free[i].size,free[i].from);
    ide_partition_put(partdata,free[i].partno);
   }
  else //remove free area information
   partition[free[i].partno]=ptype_unused; 
 for (i=1;i<partmax;i++)
  if (partition[i]==ptype_unused)
   {ide_partition_new(partdata,ptype_unused,"",0L,0L);
    ide_partition_put(partdata,i);
   }
}

int teststruct()
{//test for existing structures
 idehead=1; //starts at this head for MS-DOS compatibility
 printf("\nTesting for existing IDEDOS structures...\n");
 ide_partition_get(idedos,0);
 if (!compare(plusidedos,idedos.part_name,16))
  {
   printf(" IDEDOS structure not found on 0/%d/1 sector.\n",idehead);
   idehead=0; //or at this when no MBR
   ide_partition_get(idedos,0);
  }
 if (compare(plusidedos,idedos.part_name,16))
  {//analyse IDEDOS structure
   printf(" IDEDOS structure found on 0/%d/1 sector.\n",idehead);
   if (ide_partition_load()==0) return 1;
   is_idedos=1;
  }
 //test for MS-DOS structure
 printf("Testing for existing MS-DOS structures...\n");
 sector_read(0,0,1);
 if ((sector[0x1FE]==0x55)&&(sector[0x1FF]==0xAA)) 
  {printf(" MS-DOS structure found on 0/0/1 sector.\n");
   if (cylinders<2)
    {//cylinders, heads and sectors can be guessed from MBR
    }
   is_mbr=1;
  };
 //test if the RECORD 1 of B-DOS (Sam) is empty
 //printf("Testing for existing B-DOS structures...\n");
 //test for VDT (MB-02), others ???
 return 0;
}

int askparams()
{//Ask the user to enter disk structure
 char buf[255];
 printf("\nEnter disk parameters (0 to Quit):"
  "\n Number of cylinders:");
 fgets(buf,6,stdin);
 sscanf(buf,"%d",&cylinders);
 if (cylinders==0) return 0;
 printf("\n Number of heads:");
 fgets(buf,6,stdin);
 sscanf(buf,"%d",&heads);
 if (heads==0) return 0;
 printf("\n Number of sectors:");
 fgets(buf,6,stdin);
 sscanf(buf,"%d",&sectors);
 if (sectors==0) return 0;
 return 1;
}

int quicksystem(int entries)
{//create system partition and free area description
 int i; //for variable
 idehead=1; //when no IDEDOS, start it from head 1
 for (i=1;i<MAXPART;i++)
  partition[i]=ptype_unused;
 partition[0]=ptype_system;
 ide_partition_new(idedos,ptype_system,plusidedos,1L,1L);
 idedos.part_info.sinf.sinf_unit.un_cyls=cylinders;
 idedos.part_info.sinf.sinf_unit.un_hds=heads;
 idedos.part_info.sinf.sinf_unit.un_spt=sectors;
 idedos.part_info.sinf.sinf_unit.un_spc=sectors*heads;
 idedos.part_info.sinf.sinf_unit.un_part=entries;
 partmax=entries;
 partition[6]=ptype_free; //entry 6: boot files area
 free[0].partno=6;
 free[0].from=2;
 free[0].size=2*heads-2;
 partition[7]=ptype_free; //entry 7: data area from cylinder 2
 free[1].partno=7;
 free[1].from=2*heads;
 free[1].size=(cylinders-2)*heads;
 freecnt=2; // remember to write them to disk
 is_idedos=2; //1 - fonund on disk, 2 - virtual (not written)
 free_count(1);
}

int quickmount(int a,int nr,int t,int tlimit)
{//if a>0 ask user about free area nr
 //decrease free area and find free partition entry
 //return partition entry number or -1 when failed
 int i;
 printf("Found free area #%d of %ldkB.",
  free[nr].partno,free[nr].size*sectors/2);
 if (t>tlimit) t=tlimit; 
 if (a+1) //ask everytime
  {
   printf(" Take it?\n"
    " 0. Skip this area.\n");
   if (free[nr].size<=(long)tlimit) //only when possible
    printf(" 1. Take whole area.\n");
   if (free[nr].size>(long)t) //only when enough place
    printf(" 2. Take exactly %ldkB.\n",(long)t*sectors/2);
   while (1)
    {i=getchar();
     if (i=='0') break;
     if ((i=='1')&&(free[nr].size<=(long)tlimit))
      {t=(int)free[nr].size; break;}
     if ((i=='2')&&(free[nr].size>(long)t)) break;
    };
  }
 else
  i='1';
 if ((i=='1')||(i=='2'))
  {i=1; //find free entry
   while ((partition[i]!=ptype_unused)&&(i<partmax)) i++;
   if (i<=partmax) 
    {printf("Mounting.\n");
     ide_partition_new(partdata,ptype_unused,"",(long)t,free[nr].from);
     free[nr].size-=t;
     free[nr].from+=t;
     //to do: if (free[nr].size==0L) release the free entry
     return i;
    }
   else
    printf("Sorry, all entries are used.\n");
  }  
 return -1; 
}

int findfree(int tracks,int tmin,int tmax,int exactly)
{//find free area of size (<2GB) range from 0.5*tracks to 1.5*tracks
 //minimum tmin tracks, maximum tmax tracks
 //return: -1 - not found; otherwise partition entry number
 //!!! add "exactly" support for partitions of fixed size (movies)
 //exactly: b0=1 can be smaller, b1=1 can be bigger
 //exactly: b2=1 can be even the smallest, b3=1 can be much bigger
 //exactly: b4=1 find the smallest, b5=1 find the biggest
 u8_t frees[MAXFREE]; //table of checked free area
 int j;
 int i;
 for (i=0;i<freecnt;i++)
  if (free[i].size<=(long)tmin) //smaller than tmin?
   {frees[i]=2; //too small - don't check again
    printf("The area #%d (%ldkB) is too small.\n",
     free[i].partno,free[i].size*sectors/2);
   }
  else
   if ((free[i].size>(long)tracks/2)&&(free[i].size<(long)tracks*3/2))
    {//if between 0.8*size to 1.2*size - use the area without question
     if ((free[i].size>(long)tracks*4/5)&&(free[i].size<(long)tracks*6/5))
      return quickmount(0,i,tracks,tmax);
     else //otherwise ask the user if that amount is OK
      {j=quickmount(1,i,tracks,tmax);
       if (j>=0)
        return j;
       else
        frees[i]=3; //user refused to use this one
      }
    }
   else //in case is too small or too big - use it in next loop
    frees[i]=1;
 for (i=0;i<freecnt;i++)
  if (frees[i]==1) //in second step ask for every unasked area
   {j=quickmount(1,i,tracks,tmax);
    if (j>=0) return j;
   }
 return -1;
}

int quickCPM22(int size)
{//creates CPM22 partition of "size" MB
 // reserves 2 cylinders for system partition, makes 8 to 32MB CP/M partition
 int alloc;
 int wantedt,wantedc;
 int i;
 if (size>128) size=128; //largest partition for DSKHNDLR;
 // size + almost_one_cylinder
 wantedt=((long)size*2048+(long)heads*sectors-1)/sectors; //2048 = 1kB * 2 sectors
 wantedc=wantedt/heads; //here we have number of cylinders
 if (wantedc>255) wantedc=255; //255 is maximum number of cylinders per partition
 wantedt=wantedc*heads; //the whole number of cylinders in tracks
 printf("Searching for %ldkB free area or close.\n",(long)wantedt*sectors/2);
 if (!is_idedos) //create system partition and save it to disk
  {quicksystem(16); ide_partition_put(idedos,0); }
 if ((i=findfree(wantedt,4*heads,255*heads,0))<0) return 0;
 ide_partition_new(partdata,ptype_cpm22,"ZXVGS",0L,0L);
 //the actual number of cylinders
 wantedc=partdata.part_sysinfo.ph_ecyl-partdata.part_sysinfo.ph_scyl+1;
 //allocation unit size 2kB for smaller than 8.3MB (521B ALV table)
 wantedt=(long)wantedc*heads*sectors/2; //now size in kB
 alloc=(wantedt<=16672L)?4:8; //allocation unit in kB
 alloc=(wantedt<=8336L)?2:alloc;
 //making XDPB for the partition
 partdata.part_info.xdpb.lspt=sectors*4;
 partdata.part_info.xdpb.bsh=(alloc==2)?4:(alloc==4)?5:6;
 partdata.part_info.xdpb.blm=(alloc==2)?15:(alloc==4)?31:63;
 partdata.part_info.xdpb.exm=(alloc==2)?0:(alloc==4)?1:3;
 partdata.part_info.xdpb.dsm=wantedt/alloc-1;
 partdata.part_info.xdpb.drm=1023;
 partdata.part_info.xdpb.al0=(alloc!=8)?0xFF:0xF0;
 partdata.part_info.xdpb.al1=(alloc==2)?0xFF:0x00;
 partdata.part_info.xdpb.cks=0;
 partdata.part_info.xdpb.off=(int)partdata.part_sysinfo.ph_shd;
 partdata.part_info.xdpb.psh=0x02;
 partdata.part_info.xdpb.phm=0x03;
 partdata.part_info.xdpb.sdn=heads-1;
 partdata.part_info.xdpb.tps=wantedc;
 partdata.part_info.xdpb.pspt=sectors;
 partdata.part_info.xdpb.fsn=1;
 partdata.part_info.xdpb.pss=0x0200;
 partdata.part_info.xdpb.rwgap=partdata.part_sysinfo.ph_scyl%256;
 partdata.part_info.xdpb.fgap=partdata.part_sysinfo.ph_scyl/256;
 partdata.part_info.xdpb.mfm=0xA1;
 partdata.part_info.xdpb.auto=0xFF;
 ide_partition_put(partdata,i);
 free_count(1);
 putfrees(); //save free area info to disk
}

int convertBDOS(int rds) //last converted
{//converts rds RECORDs into IDEDOS partitions, rest is free
 int i,j; //loop variables
 char name[17];
 int reserved; //reserved sectors at start for boot and RECORDs list
 long tracks; //number of tracks used by the rds RECORDS
 reserved=(records+63)/32;
 printf("There are %d sectors reserved before RECORD 1.\n",reserved);
 tracks=((1600L*rds+reserved+sectors-1)/sectors);
 quicksystem(15+rds);
 ide_partition_put(idedos,0);
 partition[6]=ptype_free; //entry 6: free space inside RECORD 1
 free[0].partno=6;
 free[0].from=2;
 free[0].size=(1600L+reserved)/sectors-2; //is rounded down
 partition[7]=ptype_free; //entry 7: after last RECORD, from new cylinder
 i=(tracks+heads-1)/heads; //cylinder next to used area
 free[1].partno=7;
 free[1].from=i*heads;
 free[1].size=(cylinders-i)*heads;
 if (tracks%heads) // if there are free tracks before next cylinder
  {partition[5]=ptype_free; //entry 5: after last RECORD till end of cylinder 
   free[2].partno=5;
   free[2].from=tracks;
   free[2].size=heads-tracks%heads;
   freecnt=3; //was 2 after quicksystem();
  }
 free_count(1);
 printf("Converting RECORDs 2..%d.\n",rds);
 name[16]='\0'; //for display only
 for (i=1;i<rds;i++)
  {sector_read(0,0,2+i/32);
   //create partition entry
   for (j=0;j<16;j+=2) //rewerse bytes in partition name
    {name[j]=sector[i*16+j+1]; name[j+1]=sector[i*16+j]; }
//   printf("Converting RECORD %d (\"%s\").\n",i+1,name);
   ide_partition_new(partdata,ptype_record,name,
    (1599L+sectors)/sectors,(1600L*i+reserved)/sectors);
   partdata.part_sysinfo.ph_rsvd.img.vp_first=(u8_t)((1600L*i+reserved)%sectors);
   partdata.part_sysinfo.ph_rsvd.img.vp_cyls=80;
   partdata.part_sysinfo.ph_rsvd.img.vp_hds=2;
   partdata.part_sysinfo.ph_rsvd.img.vp_spt=10;
   ide_partition_put(partdata,i+15); //write changes
  }
 putfrees();
}

void operation()
{//management
 char buf[255];
 int r;
 char i=' ';
 while (i!='0')
  {printf("\nPlease select the action.\n"
    " 0. Quit program.\n"
    " 1. Quick 8MB CPM22QED/ZXVGS partition.\n"
    " 2. Quick 16MB CPM22QED/ZXVGS partition.\n"
    " 3. Quick 32MB CPM22QED/ZXVGS partition.\n");
   if (!is_idedos) //only if there's no IDEDOS already
    printf(" C. Convert B-DOS RECORDs into IDEDOS.\n");
   i=getchar();
   switch (i)
    {case '1':  quickCPM22(8); break;
     case '2':  quickCPM22(16); break;
     case '3':  quickCPM22(32); break;
     case 'C':
      if (!is_idedos)
       {r=sectors*8-16; //entries free for RECORDs
        if (r>records) r=records;
        printf("You can convert RECORDs up to %d.\n"
         " Enter last number of RECORD to convert:",r);
        fgets(buf,6,stdin);
        sscanf(buf,"%d",&r);
        if (r>1) convertBDOS(r);
       }
    }
  }
}

int main()
{cls();
 printf("The QDISK 0.32alpha freeware partition management program.\n"
  "(C) 2004-04-30 Jarek Adamski\n"
  "www.zx.yarek.pl/software/qdisk/\n"
  "                   !!! Caution !!!\n"
  "This program can seriously damage data on your disk!\n"
  "B-DOS users: make sure the RECORD 1 is empty!\n\n"
  "MAKE SURE YOU HAVE A BACKUP COPY OF THE DISK!\n\n"

  "--> Step 1. The way to access the disk.\n");
 //select DSKHNDLR (Z80) or .HDF file (stdio) or IDE devide (PC)
 if (!select()) return 0;

 printf("\n--> Step 2. The disk structure.\n");
 if (teststruct()) return 1; //try to read disk structure
 while (cylinders==1)
  {if (!askparams()) return; //ask for cylinders, heads and sectors
   if ((cylinders<40)||(heads<2)||(sectors<16)||(heads>16)||(sectors>255))
    {printf("\nWrong parameters. Please enter once again."); cylinders=1;}
  }
 records=((long)cylinders*heads*sectors+1599)/1600;
 printf("\nThe disk structure is: %d/%d/%d = %d B-DOS RECORDs.",
  cylinders,heads,sectors,records);
 printf("\n\n--> Step 3. The partition management.\n");
 operation();
}
