// this file is useable only for Z80 versions of QDISK
// it loads the drivers set into 0x8000
// then moves the selected driver to 0x8000 and calls it there

#include <string.h>

#define DSKHNDLR 0x8000
#define QDISKDAT 0x8400

void printname(unsigned char *n)
{//prints DSKHNDLR name
 char *name;
 printf("DSKHNDLR.");
 name=(char*)(n+24);
 while ((*name>' ')&&(name<n+32))
  putchar(*name++);
}

char list(void)
{//displays list of DSKHNDLRs - only for Z80 machines
 unsigned char *driver=QDISKDAT;
 char letter='@';
 loadany("QDISK.DAT",(int)driver,0x1000);
 while (driver[16]=='D')
  {letter++; //skipped for wrong drivers (e.g. FDC)
   if ((driver[6]=='\x02')&&(driver[7]=='\xA0'))
    {putchar(' ');
     putchar(letter);
     printf(". Use ");
     printname(driver);
     putchar('.\n');
    }
   driver+=driver[8]+256*(int)driver[9]; //next one
  }
 return (int)letter;
}

/* perhaps in future
int load()
{//load DSKHNDLR from file
 char buf[128];
#ifdef ZXVGS
 if (!selector("Select DSKHNDLR to load","*.*",buf)) return 0;
#else
 scanf(buf);
 if (buf=="") return 0;
#endif
 loadany(buf,DSKHNDLR,0x0400);
 return 1;
#endif
} */

int __FASTCALL__ init(int u)
{//initializes DSKHNDLR
#asm
	LD	C,L
	LD	A,0
	LD	IX,($800E)
	CALL	$8003
	LD	HL,0		;no drives found
	AND	A
	RET	NZ		;return when error
	LD	L,B		;number of units
#endasm 
}


int readsector(int u,int c,int h,int s,char* buf)
{//reads sector using DSKHNDLR
#asm
	LD	IX,0
	ADD	IX,SP
	LD	L,(IX+2)
	LD	H,(IX+3)
	LD	E,(IX+4)
	LD	B,(IX+6)
	LD	D,(IX+8)
	LD	C,(IX+10)
	LD	A,2
	LD	IX,($800E)
	CALL	$8003
	LD	L,A
	LD	H,0
#endasm 
}


int writesector(int u,int c,int h,int s,char* buf)
{//write sector using DSKHNDLR
#asm
	LD	IX,0
	ADD	IX,SP
	LD	L,(IX+2)
	LD	H,(IX+3)
	LD	E,(IX+4)
	LD	B,(IX+6)
	LD	D,(IX+8)
	LD	C,(IX+10)
	LD	A,3
	LD	IX,($800E)
	CALL	$8003
	LD	L,A
	LD	H,0
#endasm 
}

int use(int nr)
{//select driver to use: 0 means first one
 //only for Z80 machines
 unsigned char *driver=QDISKDAT; //hmmm... for non-Z80...
 int i=0;
 while (i++<nr)
  driver+=driver[8]+256*driver[9]; //next one
  
 if ((driver[6]!=2)||(driver[7]!='\xA0')) return 0;
 if ((driver[10]!=0)||(driver[11]!='\x80'))
  {printname(driver);
   printf(" has wrong destination address.\n");
   return 0;
  }
 memcpy(DSKHNDLR,driver,driver[8]+256*driver[9]); //to use area
 printf("Testing ");
 printname(driver);
 printf("... ");
 i=init(0xA0);
 printf("%d unit(s) found.\n",i);
 return i!=0;
}

int select(void)
{//select driver
 int i,j;
 char last;
 printf("\nSelect the DSKHNDLR:\n");
 printf(" 0. Quit program.\n");
// printf(" 1. Load the DSKHNDLR from a file.\n");
 last=list();
 while (1)
  {i=getchar();
   if (i=='0') break;
//   if (i=='1') if (load()) {last='@'; if (use(0)) break;}
   if (last>'@') if (i<=last) if (use(i-'A')) break;
  };
 return i!='0';
}
